# coding: utf-8

import re
from abc import ABC, abstractmethod
import ipaddress

from . import _common

_KEY_STATEMENT = _common.KEY_STATEMENT

# flags for internal process
_FLAG_UNKNOWN = 0
_FLAG_FIXED = 1
_FLAG_SEPARATORS = 2


class StatementParser:
    """Parser for statement parts in log messages.

    Statement parts in log messages describe the event in free-format text.
    This parser will segment the text into words and theire separators.
    The words are parsed as 'words' item, and the separators are
    parsed as 'symbols' item.

    The behavior of this parser is defined with a list of actions.
    The actions are sequentially applied into the statement,
    and separate it into a sequence of words (and separator symbols).

    Args:
        actions (list of any action): Segmentation rules.
            The rules are sequentially applied to the input statement.
    """

    def __init__(self, actions):
        self._l_act = actions

    @staticmethod
    def _verbose(act, input_parts):
        act_name = act.__class__.__name__
        words = []
        for part, flag in input_parts:
            if flag == _FLAG_FIXED:
                words.append("#" + part + "#")
            elif flag == _FLAG_UNKNOWN:
                words.append("'" + part + "'")
            else:
                pass
        print("{0}: {1}".format(act_name, ", ".join(words)))

    @staticmethod
    def _separate(iterable_parts):
        l_w = []
        l_s = []
        prev_isword = True
        for part, flag in iterable_parts:
            current_isword = flag in (_FLAG_FIXED, _FLAG_UNKNOWN)
            if current_isword:
                if part == "":
                    # empty word: remove
                    pass
                elif prev_isword:
                    # separator is missing: add empty separator
                    l_s.append("")
                    l_w.append(part)
                else:
                    l_w.append(part)
            else:
                if prev_isword:
                    l_s.append(part)
                else:
                    # continuous separator: merge into 1 separator
                    l_s[-1] = l_s[-1] + part
            prev_isword = current_isword
        else:
            if prev_isword:
                l_s.append("")

        assert len(l_s) == len(l_w) + 1
        return l_w, l_s

    def process_line(self, statement: str, verbose: bool = False):
        """Parse statement part of a log message (i.e., a line).

        Args:
            statement (string): String of statement part.
            verbose (bool, optional): Show intermediate progress
                of applying actions.

        Returns:
            tuple: List of two components: words and symbols.
            First component, words, is list of words.
            Second component, symbols, is list of separator string symbols.
            The length of symbols is always len(words)+1, which includes
            one before first word and one after last word.
            Some symbols can be empty string.
        """
        if verbose:
            print("Statement: {0}".format(statement))
        iterable_parts = [(statement, _FLAG_UNKNOWN), ]

        for act in self._l_act:
            iterable_parts = act.do(iterable_parts)
            if verbose:
                iterable_parts = list(iterable_parts)
                self._verbose(act, iterable_parts)
        return self._separate(iterable_parts)


class _ActionBase(ABC):

    @abstractmethod
    def do(self, input_parts):
        raise NotImplementedError

    @staticmethod
    def _is_active_part(s, flag):
        # UNKNOWN -> true (to be processed)
        # FIXED or SEPARATOR -> false (as is)
        return len(s) > 0 and flag == _FLAG_UNKNOWN

    @staticmethod
    def _standard_patterns(patterns):
        if isinstance(patterns, str):
            return [re.compile(patterns)]
        else:
            return [re.compile(p) for p in patterns]

    @staticmethod
    def _separate_multiple_match(part, iterable_mo, label_match, label_other):
        current = 0
        length = len(part)
        for mo in iterable_mo:
            if mo.start() > current:
                yield part[current:mo.start()], label_other
            yield part[mo.start():mo.end()], label_match
            current = mo.end()
        else:
            if current < length:
                yield part[current:length], label_other

    @staticmethod
    def _separate_partial_match(part, mo, match_groups, label_other):
        """Considering the given Match Object (mo),
        this function separates the given string (part)
        into three string (the second string is the matching part).
        Yields the separated string and its label (match_group name or label_other)."""
        current = 0
        length = len(part)
        sorted_match_groups = sorted(match_groups.keys(), key=lambda x: mo.start(x))
        for match_group in sorted_match_groups:
            label_match = match_groups[match_group]
            if mo.start(match_group) == -1:
                continue
            if mo.start(match_group) < current:
                raise ValueError("Invalid pattern with duplicated name groups")
            if mo.start(match_group) > current:
                yield part[current:mo.start(match_group)], label_other
            yield part[mo.start(match_group):mo.end(match_group)], label_match
            current = mo.end(match_group)
        else:
            if current < length:
                yield part[current:length], label_other


class Fix(_ActionBase):
    """Add Fixed flag to matched parts.

    Fixed parts will not be segmented by following actions.
    Fixed parts are selected by regular expression of given pattern
    (see `re <https://docs.python.org/ja/3/library/re.html>`_).

    Example:
        >>> p = StatementParser([Split(" "), Fix(r".+\\.txt"), Split(".")])
        >>> p.process_line("parsing sample.txt done.")
        (['parsing', 'sample.txt', 'done'], ['', ' ', ' ', '.'])

    Args:
        patterns (str or list of str):
            Regular expression patterns.
            If multiple patterns are given, they are matched
            with every word in order.
    """

    def __init__(self, patterns):
        self._init_patterns(patterns)

    def _init_patterns(self, patterns):
        self._l_regex = self._standard_patterns(patterns)

    def _test_match_pattern(self, s):
        for reobj in self._l_regex:
            m = reobj.match(s)
            if m:
                return True
        else:
            return False

    def do(self, iterable_parts):
        """Apply this action to every part.
        Matched parts will be fixed.
        This function works as like a filter of the statement.

        Args:
            iterable_parts (iterator of tuple): Sequence of tuples,
                including segmented statement and its annotation label.

        Yields:
            tuple: same format as the input arguments.
        """
        for s, flag in iterable_parts:
            if not self._is_active_part(s, flag):
                yield s, flag
            elif self._test_match_pattern(s):
                yield s, _FLAG_FIXED
            else:
                yield s, flag


class _PartialActionBase(_ActionBase, ABC):

    # private attributes for child classes
    _l_regex = None
    # _rest_flag = None
    _flag_other = None
    _match_groups = None
    # _fix_groups = None
    # _remove_groups = None

    def __init__(self, patterns, recursive=False):
        self._l_regex = self._init_patterns(patterns)
        self._match_groups = self._init_match_groups()
        self._recursive = recursive

    @abstractmethod
    def _init_patterns(self, patterns):
        raise NotImplementedError

    @abstractmethod
    def _init_match_groups(self):
        raise NotImplementedError

    def _separate_partial(self, part, flag, reobj):
        if self._recursive:
            return self._separate_partial_recursive(part, flag, reobj)
        else:
            mo = reobj.match(part)
            if mo:
                return self._separate_partial_match(
                    part, mo, self._match_groups, self._flag_other
                )
            else:
                return [(part, flag), ]

    def _separate_partial_recursive(self, part, flag, reobj):
        mo = reobj.match(part)
        if mo:
            # match -> separate matched part and others
            iterobj = self._separate_partial_match(
                part, mo, self._match_groups, self._flag_other
            )
            for tmp_part, tmp_flag in iterobj:
                if tmp_flag == self._flag_other:
                    iterobj = self._separate_partial_recursive(
                        tmp_part, tmp_flag, reobj
                    )
                    for p in iterobj:
                        yield p
                else:
                    yield tmp_part, tmp_flag
        else:
            yield part, flag

    @staticmethod
    def _separate_partial_match(part, mo, match_groups, label_other):
        """Considering the given Match Object (mo),
        this function separates the given string (part)
        into three string (the second string is the matching part).
        Yields the separated string and its label (match_group name or label_other)."""
        current = 0
        length = len(part)
        sorted_match_groups = sorted(match_groups.keys(), key=lambda x: mo.start(x))
        for match_group in sorted_match_groups:
            label_match = match_groups[match_group]
            if mo.start(match_group) == -1:
                continue
            if mo.start(match_group) < current:
                raise ValueError("Invalid pattern with duplicated name groups")
            if mo.start(match_group) > current:
                yield part[current:mo.start(match_group)], label_other
            yield part[mo.start(match_group):mo.end(match_group)], label_match
            current = mo.end(match_group)
        else:
            if current < length:
                yield part[current:length], label_other

    def do(self, iterable_parts):
        parts = iterable_parts
        for reobj in self._l_regex:
            new_parts = []
            for s, flag in parts:
                if self._is_active_part(s, flag):
                    for p in self._separate_partial(s, flag, reobj):
                        new_parts.append(p)
                else:
                    new_parts.append((s, flag))
            parts = new_parts

        for p in parts:
            yield p


class FixPartial(_PartialActionBase):
    """Extended Fix action to accept complicated patterns.

    Usual :class:`Fix` consider the matched part as a word, and fix it.
    In contrast, :class:`FixPartial` allow the matched part
    to include multiple fixed words or separators.

    Usecase 1:
        | e.g., :samp:`source 192.0.2.1.80 initialized.`

        If you intend to consider :samp:`192.0.2.1.80` as a combination
        of two different word: IPv4 address :samp:`192.0.2.1` and port number :samp:`80`,
        this cannot be segmented with simple Fix and Split actions.
        Following example with :class:`FixPartial` can fix these two variables.

    Example:
        >>> pattern = r'^(?P<ipaddr>(\\d{1,3}\\.){3}\\d{1,3})\\.(?P<port>\\d{1,5})$'
        >>> parser = StatementParser([Split(" "), FixPartial(pattern, fix_groups=["ipaddr", "port"]), Split(".")])
        >>> parser.process_line("source 192.0.2.1.80 initialized.")
        (['source', '192.0.2.1', '80', 'initialized'], ['', '', '.', '', '.'])

    Usecase 2:
        | e.g., :samp:`comment added: "This is a comment description".`

        If you intend to consider the comment (strings between parenthesis)
        as a word without segmentation,
        this cannot be achieved with simple Fix and Split actions.
        Following example with :class:`FixPartial` can fix the comment part.

    Example:
        >>> pattern = r'^.*?(?P<left>")(?P<fix>.+?)(?P<right>").*$'
        >>> parser = StatementParser([FixPartial(pattern, fix_groups=["fix"], \\
        ... remove_groups=["left", "right"], rest_remove=False), Split(' :.')])
        >>> parser.process_line('comment added: "This is a comment description".')
        (['comment', 'added', 'This is a comment description'], ['', ' ', ': "', '".'])

    Args:
        patterns (str): Regular expression patterns.
            If multiple patterns given, the first matched pattern
            is used to Fix the part (other patterns are ignored).
        fix_groups (str or list of str): Name groups in the patterns to fix.
            e.g., ["ipaddr", "port"] for Usecase 1.
            Unspecified groups are not fixed,
            so you can use other group names to other re functions
            like back references.
        recursive (bool, optional): If True, the patterns will be searched recursively.
        remove_groups (str or list of str, optional): Name groups in the patterns
            that should be considered as separators.
        rest_remove (bool, optional): This option determines
            how to handle strings outside the fixed groups.
            e.g., 'comment added: "' and '".' in Usecase 2.
            Defaults to False, which means they are left as parts
            for further actions.
            In contrast, if True, they are considered as separators
            and will not be segmented or fixed further.
    """

    def __init__(self, patterns, fix_groups,
                 recursive=False, remove_groups=None, rest_remove=False):
        if isinstance(fix_groups, str):
            self._fix_groups = [fix_groups]
        else:
            self._fix_groups = fix_groups

        if remove_groups is None:
            self._remove_groups = []
        elif isinstance(remove_groups, str):
            self._remove_groups = [remove_groups]
        else:
            self._remove_groups = remove_groups

        if rest_remove:
            self._flag_other = _FLAG_SEPARATORS
        else:
            self._flag_other = _FLAG_UNKNOWN

        super().__init__(patterns, recursive=recursive)

    def _init_patterns(self, patterns):
        return self._standard_patterns(patterns)

    def _init_match_groups(self):
        match_groups = {}
        match_groups.update({gname: _FLAG_FIXED
                             for gname in self._fix_groups})
        match_groups.update({gname: _FLAG_SEPARATORS
                             for gname in self._remove_groups})
        return match_groups


class FixParenthesis(_PartialActionBase):
    """Extended FixPartial easily used to fix strings between parenthesis.

    The basic usage is similar to FixPartial, but
    this class is designed especially for parenthesis,
    and the format of patterns is simpler.
    For example, FixParenthesis with pattern :samp:`['"', '"']` work samely as
    FixPartial with pattern :regexp:`r'^.*?"(?P<fix>.+?)".*$'`.

    Each pattern is a 2-length list of left and right parenthesis.
    The left and right pattern can consist of multiple characters,
    such as :samp:`["<!--", "-->"]`.

    Example:
        >>> parser = StatementParser([FixParenthesis(['"', '"']), Split(' .:"')])
        >>> parser.process_line('comment added: "This is a comment description".')
        (['comment', 'added', 'This is a comment description'], ['', ' ', ': "', '".'])

    Note: If a statement has multiple pairs of parenthesis,
    you need to add multiple FixParenthesis action to StatementParser actions.
    This is because FixParenthesis accept only one fix_group
    to extract in the action.
    """
    _key_fix = "fix"
    _fix_groups = [_key_fix, ]
    _key_left = "left"
    _key_right = "right"
    _remove_groups = [_key_left, _key_right]
    _flag_other = _FLAG_UNKNOWN

    def _init_patterns(self, patterns):
        if isinstance(patterns, str):
            return [self._init_named_pattern(patterns)]
        elif len(patterns) == 2 and isinstance(patterns[0], str):
            return [self._init_named_pattern(patterns)]
        else:
            return [self._init_named_pattern(pattern)
                    for pattern in patterns]

    @classmethod
    def _init_named_pattern(cls, parent_pattern):
        # non-greedy match
        assert len(parent_pattern) == 2, "Invalid pattern for FixParenthesis"
        p_left = parent_pattern[0]
        p_right = parent_pattern[1]
        restr = r'^.*?' + \
                r'(?P<' + cls._key_left + '>' + re.escape(p_left) + ')' + \
                r'(?P<' + cls._key_fix + r'>.+?)' + \
                r'(?P<' + cls._key_right + '>' + re.escape(p_right) + r')' + \
                r'.*$'
        return re.compile(restr)

    def _init_match_groups(self):
        match_groups = {}
        match_groups.update({gname: _FLAG_FIXED
                             for gname in self._fix_groups})
        match_groups.update({gname: _FLAG_SEPARATORS
                             for gname in self._remove_groups})
        return match_groups


class FixIP(_ActionBase):
    """Add Fixed flag to the parts of IP addresses.

    This class use ipaddress library instead of regular expression.

    Args:
        address: match IP addresses, defaults to True
        network: match IP networks, defaults to True
    """

    def __init__(self, address=True, network=True):
        super().__init__()
        self._test_addr = address
        self._test_net = network

    @staticmethod
    def _is_ipaddr(string, ipaddr=True, ipnet=True):
        # fast check: no digit char, no ipaddr
        # because ipaddress lib is slow
        if "." not in string and ":" not in string:
            return False

        if ipaddr:
            try:
                ipaddress.ip_address(string)
            except ValueError:
                pass
            else:
                return True
        if ipnet and "/" in string:
            try:
                ipaddress.ip_network(string, strict=False)
            except ValueError:
                pass
            else:
                return True
        return False

    def _test_match_pattern(self, s):
        return self._is_ipaddr(s, self._test_addr, self._test_net)

    def do(self, iterable_parts):
        """same as Fix.do"""
        for s, flag in iterable_parts:
            if not self._is_active_part(s, flag):
                # FIXED or SEPARATOR -> as is
                yield s, flag
            elif self._test_match_pattern(s):
                # match -> flag changed to FIXED
                yield s, _FLAG_FIXED
            else:
                # not match -> as is
                yield s, flag


class Remove(_ActionBase):
    """Add Separator flag to matched parts.

    Separator parts will be ignored by following actions.
    Separator parts are selected by regular expression of given pattern
    (see `re <https://docs.python.org/ja/3/library/re.html>`_).

    Args:
        patterns (str or list of str):
            Regular expression patterns.
            If multiple patterns are given, they are matched
            with every word in order.
    """

    def __init__(self, patterns):
        self._init_patterns(patterns)

    def _init_patterns(self, patterns):
        self._l_regex = self._standard_patterns(patterns)

    def _test_match_pattern(self, s):
        for reobj in self._l_regex:
            m = reobj.match(s)
            if m:
                return True
        else:
            return False

    def do(self, iterable_parts):
        """Apply this action to every part.
        Matched parts will be fixed.
        This function works as like a filter of the statement.

        Args:
            iterable_parts (iterator of tuple): Sequence of tuples,
                including segmented statement and its annotation label.

        Yields:
            tuple: same format as the input arguments.
        """
        for s, flag in iterable_parts:
            if not self._is_active_part(s, flag):
                yield s, flag
            elif self._test_match_pattern(s):
                yield s, _FLAG_SEPARATORS
            else:
                yield s, flag


class RemovePartial(_PartialActionBase):
    """Extended Remove action to accept complicated patterns.

    Usual :class:`Remove` consider the matched part as a separator.
    In contrast, :class:`RemovePartial` allow
    partially removing separators from a word matching with the given patterns.

    Example:
        >>> rpattern = r'^.*([^:](?P<colon>:))$'
        >>> fpattern = r'^\\d{2}:\\d{2}:\\d{2}\\.\\d{3}$'
        >>> rules = [Split(" "), RemovePartial(rpattern, remove_groups=["colon"]), Fix(fpattern), Split(":")]
        >>> parser = StatementParser(rules)
        >>> parser.process_line("2000 Mar 4 12:34:56.789: message: duplicated header")
        (['2000', 'Mar', '4', '12:34:56.789', 'duplicated', 'header'], ['', ' ', ' ', ' ', ': ', ' ', ''])
    """

    _flag_other = _FLAG_UNKNOWN

    def __init__(self, patterns, remove_groups, recursive=False):
        if isinstance(remove_groups, str):
            self._remove_groups = [remove_groups]
        else:
            self._remove_groups = remove_groups

        super().__init__(patterns, recursive=recursive)

    def _init_patterns(self, patterns):
        return self._standard_patterns(patterns)

    def _init_match_groups(self):
        match_groups = {}
        match_groups.update({gname: _FLAG_SEPARATORS
                             for gname in self._remove_groups})
        return match_groups


class Split(_ActionBase):
    """Split statement (or its parts) by given separators.

    For example, separators of white space and dot translates

    | :samp:`['This is a statement.'] -> ['This', 'is', 'a', 'statement']`

    The removed separators (white space and dot in this case)
    will not be considered in further actions.

    Example:
        >>> parser = StatementParser([Split(" .")])
        >>> parser.process_line("This is a statement.")
        (['This', 'is', 'a', 'statement'], ['', ' ', ' ', ' ', '.'])

    Args:
        separators (str): separator symbol strings.
            If iterable, they imply joined and used all for segmentation.
            Escape sequence is internally added, so you don't need
            to add it manually.
    """

    def __init__(self, separators):
        if not isinstance(separators, str):
            separators = "".join(separators)
        restr = r'([' + re.escape(separators) + '])+'
        self._regex = re.compile(restr)

    @staticmethod
    def _iter_part(s, iterable_mo):
        current = 0
        length = len(s)
        for mo in iterable_mo:
            if mo.start() > current:
                yield s[current:mo.start()], _FLAG_UNKNOWN
            yield s[mo.start():mo.end()], _FLAG_SEPARATORS
            current = mo.end()
        else:
            if current < length:
                yield s[current:length], _FLAG_UNKNOWN

    def do(self, iterable_parts):
        for s, flag in iterable_parts:
            if self._is_active_part(s, flag):
                matchobjs = self._regex.finditer(s)
                for tmp_parts in self._iter_part(s, matchobjs):
                    yield tmp_parts
            else:
                yield s, flag


class ConditionalSplit(Fix, Split):
    """Split parts matching the given patterns by given separators.

    Example:
        >>> parser = StatementParser([
        >>>     Split(" ()"),
        >>>     RemovePartial(r'^.*[^:](?P<colon>:)$', remove_groups=["colon"]),
        >>>     ConditionalSplit(r'^%[A-Z]+-\\d+(-[A-Z]+-\\d+)?$', r'%-')
        >>> ])
        >>> parser.process_line("%KERNEL-4-EVENT-7: host h1-i2.example.org scored -0.035 value (20.0%)")
        ['KERNEL', '4', 'EVENT', '7', 'host', 'h1-i2.example.org', 'scored', '-0.035', 'value', '20.0%']

    Args:
        patterns (str): Regular expression patterns.
            If multiple patterns given, this action will split
            parts matching at least one of them.
        separators (str): separator symbol strings.
            If iterable, they imply joined and used all for segmentation.
            Escape sequence is internally added, so you don't need
            to add it manually.
    """

    def __init__(self, patterns, separators):
        Fix.__init__(self, patterns)
        Split.__init__(self, separators)

    def do(self, iterable_parts):
        for s, flag in iterable_parts:
            if not self._is_active_part(s, flag):
                yield s, flag
            elif self._test_match_pattern(s):
                matchobjs = self._regex.finditer(s)
                for tmp_parts in self._iter_part(s, matchobjs):
                    yield tmp_parts
            else:
                yield s, flag
